/*

    GDS7 v0.1 interactive digital image transfer software for DS-7 camera
    Copyright (C) 1998 Matthew G. Martin

    Camera transfer code transcribed from a Perl Language library
    Copyright (C) 1997 Mamoru Ohno

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


*/
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <termios.h>
#include <stdio.h>
#include <strings.h>
#include <tiff.h>
#include <tiffio.h>
#include "ds7lib.h"

struct termios tiop;
int portp;
int go_ahead;

cam_info this_cam_info={
  80,60,NULL
};
int debug=0;
int lock=0; /* is camera busy */

#define ACK 0x6
#define ETX 0x3
#define ETB 0x17

/*
  Return camera info
*/
cam_info* get_cam_info(){
  return(&this_cam_info);
};


/* 
   Wait for an acknowlege from the camera
 */

int ack(int iop){
  char buffer[1];
  int timeout;

  timeout=100;
  while((read(iop,buffer,1)==0) && timeout){
    usleep(100);
    //    printf("timeout=%d\n");
    timeout--;
  };
  if(!timeout) printf("Ack timeout\n");
  if (buffer[0]!=ACK) printf("get %X instead of ACK\n",buffer[0]);
  return(! timeout || buffer[0]!=ACK);
};

int ok(int iop){
  write(iop,"\x6",1);
};

int make_msg(char* inbuf,int len,char* fmtbuf){
  int i,fmtp,bcc;

  fmtbuf[0]=0x10;
  fmtbuf[1]=0x02;
  fmtp=2;
  bcc=3;

  for (i=0;i<len;i++){
    fmtbuf[fmtp++]=inbuf[i];
    /* hex 10 is escaped by itself */
    if (inbuf[i]==0x10) fmtbuf[fmtp++]=0x10;
    bcc^=inbuf[i];
  };
  fmtbuf[fmtp++]=0x10;
  fmtbuf[fmtp++]=0x03;
  fmtbuf[fmtp++]=bcc;
  return(fmtp);
};


void serial_speed(int spdflg){
  tiop.c_cflag = spdflg | PARENB  | CLOCAL | CS8 |  CREAD;
  tcflush(portp, TCIFLUSH);
  tcsetattr(portp,TCSANOW,&tiop);
}

int open_serial(char* dev){
  int i;

  portp=open(dev,O_RDWR | O_NOCTTY);
  if (portp==-1) return(portp);
  tcgetattr(portp,&tiop);
  tiop.c_cflag = B9600 | PARENB  | CLOCAL | CS8 |  CREAD;
  tiop.c_lflag = ~ICANON;
  tiop.c_iflag = 0;
  tiop.c_oflag = 0;
  for (i=0;i<NCCS;i++) tiop.c_cc[i]=0;
  tcsetattr(portp,TCSANOW,&tiop);
  tcflush(portp, TCIFLUSH);
  return(portp);
};

/* read_serial
    Reads given "num" chars from serial port "iop" unless there
    is a timeout.
*/
int read_serial(int iop,char *buf,int num){
  int count,to,got;
  count=0;
  to=100;

  while ((count<num)&&to){
    got=read(iop,buf+count,num-count);
    if (got) {
      count+=got;
      to=100;
    }
    else to--;
    usleep(1000);
  };
  if (!to) fprintf(stderr,"read_serial timeout\n");
  return(!to);
};

int read_ds7(int iop,char *buffer,int *ctrl){
  int got,timeout,bcc;
  long rcount,count,i;
  char tmpbuf[10000];

  do{
    count=0;
    timeout=100;
    read_serial(iop,tmpbuf,2);
    if (debug)
      printf("read preamble is %X,%X,%X,%X\n"
	     ,tmpbuf[0],tmpbuf[1],tmpbuf[2],tmpbuf[3]);
    do{
      usleep(10000);
      got=read(iop,tmpbuf+count,1024);
      if (debug>1) printf("Got %d bytes\n",got);
      if (got==0) timeout--;
      else timeout=100;
      count+=got;
    } while (timeout&&((count<5)||!((tmpbuf[count-3]==0x10)
				    &&(tmpbuf[count-2]!=0x10)&&(tmpbuf[count-4]!=0x10))));

    *ctrl=tmpbuf[count-2];

    rcount=bcc=0; 
    bcc^=*ctrl;
    for(i=0;i<count-3;i++) {
      if (i>3) buffer[rcount++]=tmpbuf[i];
      bcc^=tmpbuf[i];
      if (tmpbuf[i]==0x10) {
	i++;
	if (tmpbuf[i]!=0x10) fprintf(stderr,"Unmatched escape!\n");
      };
    };
    bcc^=tmpbuf[count-1];
    if (bcc!=0) {
      printf("Checksum Error\n");
      write(iop,"\x15",1);
    };
  }while (bcc!=0);

  if (debug) printf("Got %d bytes, ctrl=%X\n",count-3,tmpbuf[count-3]);
  if (debug) printf("Final Bytes are %X,%X,%X,%X\n",tmpbuf[count-4],
		      tmpbuf[count-3],tmpbuf[count-2],tmpbuf[count-1]);
  return(rcount);
};

/*
  Opens communications with the DS7
   gets number of photos available and camera ID string
*/

int open_ds7(char* devname){
  char buffer[20];
  int cmd;

  if ((portp=open_serial(devname))==-1) {
    fprintf(stderr,"Error opening \"devname\"");
    return(portp);
  };

  write(portp,"\x5",1);
  if (ack(portp)){
    /* Try sending shutdown at each baudrate */
    serial_speed(B57600);
    write(portp,"\x4",1);
    tcdrain(portp);
    serial_speed(B38400);
    write(portp,"\x4",1);
    tcdrain(portp);
    serial_speed(B19200);
    write(portp,"\x4",1);
    tcdrain(portp);
    serial_speed(B9600);
    usleep(50000);
    write(portp,"\x5",1);
    if (ack(portp))return(0);
  };

  write(portp,buffer,make_msg("\0\x9\0\0",4,buffer));
  if (ack(portp)) return(0);

  read_ds7(portp,buffer,&cmd);
  ok(portp);
  buffer[9]='\0'; /* Just in case*/
  if (this_cam_info.cam_id!=NULL) free(this_cam_info.cam_id);
  this_cam_info.cam_id=strdup(buffer);
  //  printf("Camera Is %.9s\n",this_cam_info.cam_id);
  
  write(portp,buffer,make_msg("\0\xb\0\0",4,buffer));
  if (ack(portp)) return(0);

  read_ds7(portp,buffer,&cmd);
  ok(portp);
  //  printf("%d images\n",buffer[0]);
  portp=portp;
    return(buffer[0]);
  //  return(portp);
};

int set_speed(int speed){
  char cmd[5]="\x1\x7\x1\0\x7";
  char buffer[20];
  int dummy,spd_flg;

  if (speed<=9600){
    cmd[4]=0;
    spd_flg=B9600;
  }
  else if (speed<=19200){
    cmd[4]=4;
    spd_flg=B19200;
  }
  else if (speed<=38400){
    cmd[4]=6;
    spd_flg=B38400;
  }
  else{
    cmd[4]=7;
    spd_flg=B57600;
  };

  write(portp,buffer,make_msg(cmd,5,buffer));
  if (ack(portp)) return(1);
  read_ds7(portp,buffer,&dummy);
  ok(portp);
  tcdrain(portp);

  write(portp,"\x4",1);
  tcdrain(portp);

  usleep(200000);
  serial_speed(spd_flg);

  write(portp,"\x5",1);

  if (ack(portp)) return(1);
  ok(portp);
  return(0);
};

int image_name(int num,char *name){
  char cmd[6]="\0\xa\x2\0\0\0";
  char buffer[20];
  int dummy;

  cmd[4]=num;
  write(portp,buffer,make_msg(cmd,6,buffer));
  if (ack(portp)) return(0);

  read_ds7(portp,buffer,&dummy);
  ok(portp);
  strncpy(name,buffer,12);
  name[12]=0;
};

int image_date(){
};

char* get_info(int picnum){
  char* message;
  message=(char *) malloc(100);
  sprintf(message,"Info for Picture %d\n\nNot yet Implemented\n",picnum);
  return(message);
};

long get_data(int num,int thumb,char * buffer,void progcb(float percent),long max){
  char cmd[6]="\0\x2\x2\0\0\0";
  int ctrl,imgptr,got;

  go_ahead=1;
  if (lock) return(0);
  lock=1;

  cmd[4]=num;
  if (thumb) cmd[1]=0;

  write(portp,buffer,make_msg(cmd,6,buffer));
  if (ack(portp)) {
    lock=0;
    return(0);
  };
  imgptr=0;
  do{
    got=read_ds7(portp,buffer+imgptr,&ctrl);
    imgptr+=got;
    if (debug) printf("Got %d bytes\n",got);
    if(ctrl==ETB) {
      fflush(stdout);
      ok(portp);
      /* Got a packet, tell the world */
      progcb(1.0*imgptr/max);
    };
    if (debug) printf("%0.2X",ctrl);
    fflush(stdout);
  } while((ctrl!=ETX)&&(go_ahead));
  if (!go_ahead) {
    write(portp,"\x0",1);  /* Find interrupt code */
  }
  else ok(portp);
  //  printf("Image has %d bytes\n",imgptr);

  /* All done, Set the progress callback to zero */
  progcb(0);
  lock=0;
  if (!go_ahead) imgptr=0;
  return(imgptr);
};

int close_ds7(){
  ok(portp);
  write(portp,"\x4",1);
  close(portp);
};

long lilend(u_char *data, int size){
  long total;

  total=0;
  for (--size;size>=0;size--){
    total<<=8;
    total+=data[size];
  };

  return(total);
};

/* directly from Ohno's perl script (IFD=Image Format Descriptor ?)*/

long next_ifd(u_char *exif,int num){
  int offset=(exif[num]+(exif[num+1]<<8))*12+num+2;
  //  printf("nifd,offset=%ld\n",offset);
  return(lilend(exif+offset,4));
};

/* 
   Pull value of tag tagnum from data starting at offset 
      only works for 4 byte values (for now)

*/

int getval(u_char *data, int offset, int tagnum){
  int numtags,i,tag;

  numtags=lilend(data+offset,2);
  // printf("getval:%d tags\n",numtags);

 i=-1;
 do{
   i++;
   tag=lilend(data+i*12+offset+2,2);
 }while((i<numtags)&&(tag!=tagnum));

 if(tag!=tagnum) {
   printf("Tag %d not found\n",tagnum);
   exit(1);
 };

 return(lilend(data+i*12+offset+10,4));
};

int sizetab[13]={
  0,1,1,2,4,8,1,1,2,4,8,4,8
};

/* 
   Return tag number of directory entry at tagind
 */
int tagnum(u_char *data,int offset,int tagind){
 return(lilend(data+tagind*12+offset+2,2));
};

/* 
   Return data size of directory entry at tagind
 */
int datsize(u_char *data,int offset,int tagind){
  return(sizetab[lilend(data+tagind*12+offset+4,2)]*lilend(data+tagind*12+offset+6,4));
};

/* 
   Return "value" of directory entry at tagind
 */
int theval(u_char *data,int offset,int tagind){
 return(lilend(data+tagind*12+offset+10,4));
};

/* 
   Set the "value" of directory entry at tagind
*/

int setval(u_char *data,int offset,int tagind,long newval){
  int i;
  for (i=0;i<4;i++) data[tagind*12+offset+10+i]=0xff&(newval>>(i*8));
  if (theval(data,offset,tagind)!=newval) printf("Setptr: error %d inst %d\n",theval(data,offset,tagind),newval);
};

int get_pic(int pic,char* fname,void progcb(float p)){
  long size;
  u_char tmpbuff[100000];
  FILE *fd;

  size=get_data(pic,0,tmpbuff,progcb,65000);
  if (!size) return(size);
  
  fd=fopen(fname,"w");
  fwrite(tmpbuff,size,1,fd);
  fclose(fd);
  return(size);
};

int get_thumb(int pic,char* fname,void progcb(float p)){
  u_char tmpbuff[20000];
  u_char *imagedata;
  u_char* exif;
  int i,j,tag;
  char* thumbdata;
  unsigned int exiflen,entry;
  long offset;           /* $A in Ohno's script */
  long size,dataptr,datvec,dsize;
  FILE *fd;

  size=get_data(pic,1,tmpbuff,progcb,10500);
  if (!size) return (size); /* An Error Ocurred*/

  exif=tmpbuff+12;

  fd=fopen(fname,"w");
  /* Copy header*/
  fwrite(exif,8,1,fd);

  exiflen=tmpbuff[5]+(tmpbuff[4]<<8)-8;
  //  printf("Exif header is %X,%X=%d bytes\n",tmpbuff[4],tmpbuff[5],exiflen);

  offset=lilend(exif+4,4);
  //  printf("offset is %ld\n",offset);

  /* Skip to TIFF image data */
  offset=next_ifd(exif,/*lilend(exif+4,4)*/offset);
  if (offset==0) return 0;
  //  printf("offset is %ld\n",offset);

  /* Copy number of entries */
  fwrite(exif+offset,2,1,fd);

  entry=lilend(exif+offset,2);

  imagedata=exif+getval(exif,offset,273); /*imagedata start*/
  dataptr=getval(exif,offset,279);        /* imagedata size */

  for (i=0;i<entry;i++){
    dsize=datsize(exif,offset,i);
    tag=tagnum(exif,offset,i);
    //    printf("Datsize %d (tag=%d) is %d\n",i,tag,dsize);
    if (tag==273) {
      setval(exif,offset,i,12*entry+14); /* set to end of directory */
      fwrite(exif+offset+12*i+2,12,1,fd);
    }
    else {
      if (dsize<5)
	/* Just copy the field if small */
	fwrite(exif+offset+12*i+2,12,1,fd);
      else{
	datvec=theval(exif,offset,i);
	setval(exif,offset,i,dataptr+12*entry+14);
	for (j=0;j<dsize;j++) imagedata[dataptr++]=exif[datvec+j];
	fwrite(exif+offset+12*i+2,12,1,fd);
      };
    };
  };
  fwrite(exif+entry*12+10,4,1,fd);  /* Write 4 zero bytes */
  fwrite(imagedata,dataptr,1,fd);
  fclose(fd);
  return size;
};

void req_cancel(){
  go_ahead=0;
};

int get_status(){
  return(go_ahead);
};

void clear_status(){
  go_ahead=1;
};

